/*
-----------------------------------------------------------------------------
OgreTootle :Mesh Optimizer using ATI Tootle 1.0

Copyright (c) 2000-2006 Ahmed Ismaiel Zakaria

This program is open source; you can redistribute it and/or modify it 
or even burn it if you want :).

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.

-----------------------------------------------------------------------------
*/

#include "option.h"
#include "tootlelib.h"
#include <vector>
#include <list>
#include <string.h>
#include <string>
#include <iostream>
#include <fstream>
#define OGRE_DEBUG_MEMORY_MANAGER 0
#include "Ogre.h"
#include "OgreMeshSerializer.h"
#include "OgreSkeletonSerializer.h"
#include "OgreDefaultHardwareBufferManager.h"
#include "OgreHardwareVertexBuffer.h"
#include <sys/stat.h>

using namespace std;
using namespace Ogre;


// OgreGlobals :
//NB some of these are not directly used, but are required to
// instantiate the singletons used in the dlls
LogManager* logMgr;
Math* mth;
MaterialManager* matMgr;
SkeletonManager* skelMgr;
MeshSerializer* meshSerializer;
SkeletonSerializer* skeletonSerializer;
DefaultHardwareBufferManager *bufferManager;
ResourceGroupManager* rgm;
MeshManager* meshMgr;

//=================================================================================================================================
/// A simple structure to store the settings
//=================================================================================================================================
struct TootleSettings
{
	const char *pMeshName ;
	const char *pViewpointName ;
	unsigned int nClustering ;
	unsigned int nCacheSize;
	TootleFaceWinding eWinding;
	const char* mOutFile;
};


//=================================================================================================================================
/// A simple structure to hold Tootle statistics
//=================================================================================================================================
struct TootleStats
{
	unsigned int nClusters;
	float        fVCacheIn;
	float        fVCacheOut;
	float        fOverdrawIn;
	float        fOverdrawOut;
	float        fMaxOverdrawIn;
	float        fMaxOverdrawOut;
};



//=================================================================================================================================
/// Displays usage instructions, and calls exit()
/// \param nRet  Return code to pass to exit
//=================================================================================================================================
void ShowHelpAndExit( int nRet )  
{
	fprintf(stderr,
		"Syntax:\n"
		"   OgreTootle [-c clusters] [-s cachesize] [-f] in.mesh > out.mesh\n"
		"        If -f is specified, counter-clockwise faces are front facing.\n");
	exit( nRet );
}


//=================================================================================================================================
/// Displays a tootle error message
/// \param result The tootle result that is being displayed
//=================================================================================================================================
void DisplayTootleErrorMessage( TootleResult eResult )
{
	/*	std::cerr << "Tootle returned error: " << std::endl;
	switch( eResult )
	{
	case TOOTLE_INVALID_ARGS:     std::cerr << " TOOTLE_INVALID_ARGS"    << std::endl; break; 
	case TOOTLE_OUT_OF_MEMORY:    std::cerr << " TOOTLE_OUT_OF_MEMORY"   << std::endl; break;      
	case TOOTLE_3D_API_ERROR:     std::cerr << " TOOTLE_3D_API_ERROR"    << std::endl; break;      
	case TOOTLE_INTERNAL_ERROR:   std::cerr << " TOOTLE_INTERNAL_ERROR"  << std::endl; break;  
	case TOOTLE_NOT_INITIALIZED:  std::cerr << " TOOTLE_NOT_INITIALIZED" << std::endl; break;   
	}
	*/
}



//=================================================================================================================================
/// Displays usage instructions, and calls exit()
/// \param nRet  Return code to pass to exit
//=================================================================================================================================
void ParseCommandLine(int argc, char *argv[], TootleSettings* pSettings ) 
{
	Option::Definition options[] = {
		{ 'h', "Help" },
		{ 'c', "Number of clusters" },
		{ 's', "Post TnL vcache size" },
		{ 'f', "Treat counter-clockwise faces as front facing (instead clockwise faces)." },
		{ 'o', "Output Mesh File Name" },
		{ 0, NULL },
	};

	Option opt;
	char cOption = opt.Parse(argc, argv, options);

	while( cOption != -1 )
	{
		switch ( cOption ) 
		{
		case 'h':
			ShowHelpAndExit(0);
			break;

		case 'f':
			pSettings->eWinding = TOOTLE_CCW;
			break;

		case 'c':
			pSettings->nClustering = atoi(opt.GetArgument(argc, argv));
			break;

		case 'o':
			pSettings->mOutFile = opt.GetArgument(argc, argv);
			break;

		case 's':
			pSettings->nCacheSize = atoi(opt.GetArgument(argc, argv));
			break;

			// option with no switch.  This is the mesh filename
		case '?':
			if (!pSettings->pMeshName) 
			{
				pSettings->pMeshName = opt.GetArgument(argc, argv);
			}
			else 
			{
				ShowHelpAndExit(1);
			}
			break;

		default:
			ShowHelpAndExit(1);
			break;
		}

		cOption = opt.Parse(argc, argv, options);
	}

	// make sure we got a mesh name
	if (!pSettings->pMeshName) 
	{
		ShowHelpAndExit(1); 
	}
}


//=================================================================================================================================
/// A helper function to print formatted TOOTLE statistics
/// \param f      A file to print the statistics to
/// \param pStats The statistics to be printed
//=================================================================================================================================

void PrintStats(TootleStats* pStats)
{
	char buf[1000];
	sprintf(buf, "#Tootle Stats\n#Clusters: %u\n#CacheIn/Out: %.3fx (%.3f/%.3f)\n#OverdrawIn/Out: %.3fx (%.3f/%.3f)\n#OverdrawMaxIn/Out: %.3fx (%.3f/%.3f)\n", 
		pStats->nClusters,
		pStats->fVCacheIn/pStats->fVCacheOut, 
		pStats->fVCacheIn, 
		pStats->fVCacheOut,
		pStats->fOverdrawIn/pStats->fOverdrawOut, 
		pStats->fOverdrawIn, 
		pStats->fOverdrawOut,
		pStats->fMaxOverdrawIn/pStats->fMaxOverdrawOut, 
		pStats->fMaxOverdrawIn, 
		pStats->fMaxOverdrawOut);

	logMgr->getDefaultLog()->logMessage(buf);
}


//=================================================================================================================================
/// A helper function to retrieve vertex and index buffer from Ogre Correspondence
/// \param indexBuffer			shared pointer to mesh index buffer
/// \param vertexDeclaration	pointer to vertex buffer declaration
/// \param vertexBufferBinding	pointer to vertex buffer bindings
/// \param vertices				output vector for inserted vertices
/// \param indices				output vector for inserted indices
//=================================================================================================================================
void FillMeshData(HardwareIndexBufferSharedPtr indexBuffer,
				  VertexDeclaration *vertexDeclaration,
				  VertexBufferBinding* vertexBufferBinding,
				  vector<Vector3> & vertices,
				  vector<unsigned int> &indices)
{
	// Get the target element
	const VertexElement* srcElem = vertexDeclaration->findElementBySemantic(
		VES_POSITION);

	if (!srcElem || srcElem->getType() != VET_FLOAT3)
	{
		OGRE_EXCEPT(Exception::ERR_INVALIDPARAMS,
			"SubMesh has no position semantic!! ",
			"FillMeshData");
	}
	HardwareVertexBufferSharedPtr srcBuf;
	srcBuf = vertexBufferBinding->getBuffer(srcElem->getSource());
	unsigned char *pSrcBase;
	// lock source for read only
	pSrcBase = static_cast<unsigned char*>(
		srcBuf->lock(HardwareBuffer::HBL_READ_ONLY));
	size_t numvertices=srcBuf->getNumVertices();
	size_t vertexsize=srcBuf->getVertexSize();
	float	*pVPos;	                // vertex position buffer, read only

	for(size_t i=0;i<numvertices;i++)
	{
		srcElem->baseVertexPointerToElement(pSrcBase, &pVPos);
		vertices.push_back(Vector3(pVPos[0],pVPos[1],pVPos[2]));
		pSrcBase+=vertexsize;
	}
	srcBuf->unlock();

	//fill the index buffer
	//tootle only work with 32Bit buffers 
	//,so we'll had to recompress them later before saving

	// retrieve buffer pointers
	uint16	*pVIndices16 = NULL;    // the face indices buffer, read only
	uint32	*pVIndices32 = NULL;    // the face indices buffer, read only
	bool use32bit = false;

	if (indexBuffer->getType() == HardwareIndexBuffer::IT_32BIT)
	{
		pVIndices32 = static_cast<uint32*>(
			indexBuffer->lock(HardwareBuffer::HBL_READ_ONLY));
		use32bit = true;
		for(size_t i=0;i<indexBuffer->getNumIndexes();i++)
		{
			indices.push_back((unsigned int)(*pVIndices32));
			pVIndices32++;
		}

	}
	else
	{
		pVIndices16 = static_cast<uint16*>(
			indexBuffer->lock(HardwareBuffer::HBL_READ_ONLY));
		size_t nm_indices=indexBuffer->getNumIndexes();
		for(size_t i=0;i<nm_indices;i++)
		{
			indices.push_back((unsigned int)(*pVIndices16));
			pVIndices16++;
		}
	}
	indexBuffer->unlock();
	int a=0;

}

int main( int argc, char* argv[] )
{
	logMgr = new LogManager();
	logMgr->createLog("OgreTootle.log", true);
	rgm = new ResourceGroupManager();
	mth = new Math();
	matMgr = new MaterialManager();
	matMgr->initialise();
	skelMgr = new SkeletonManager();
	meshSerializer = new MeshSerializer();
	skeletonSerializer = new SkeletonSerializer();
	bufferManager = new DefaultHardwareBufferManager(); // needed because we don't have a rendersystem
	meshMgr = new MeshManager();


	// initialize settings to defaults
	TootleSettings settings;
	settings.pMeshName = NULL;
	settings.pViewpointName = NULL;
	settings.nClustering = 0;
	settings.nCacheSize = TOOTLE_DEFAULT_VCACHE_SIZE;
	settings.eWinding = TOOTLE_CW;
	settings.mOutFile=NULL;

	// parse the command line
	ParseCommandLine( argc, argv, &settings );

	std::vector<Vector3> vertices;
	std::vector<unsigned int> indices;

	// ***************************************************
	//   Load the mesh
	// ***************************************************

	// read the mesh from the Mesh file
	struct stat tagStat;

	FILE* pFile = fopen( settings.pMeshName, "rb" );
	if (!pFile)
	{
		OGRE_EXCEPT(Exception::ERR_FILE_NOT_FOUND, 
			"File " + String(settings.pMeshName) + " not found.", "OgreTootle");
	}
	stat( settings.pMeshName, &tagStat );
	MemoryDataStream* memstream = new MemoryDataStream(settings.pMeshName, tagStat.st_size, true);
	fread( (void*)memstream->getPtr(), tagStat.st_size, 1, pFile );
	fclose( pFile );

	Mesh mesh(meshMgr, "conversion", 0, ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);

	DataStreamPtr stream(memstream);

	MeshSerializer mMeshSerializer;
	mMeshSerializer.importMesh(stream, &mesh);
	int num_lods=mesh.getNumLodLevels();
	int num_submeshes=mesh.getNumSubMeshes();
	for(int i=0;i<num_submeshes;i++)
	{
		// build buffers containing only the vertex positions and indices, since this is what Tootle requires
		SubMesh * smesh=mesh.getSubMesh(i);
		if(smesh->operationType!=RenderOperation::OT_TRIANGLE_LIST)
		{
			continue;
		}
		if(smesh->useSharedVertices)
		{
			FillMeshData(smesh->indexData->indexBuffer,
				mesh.sharedVertexData->vertexDeclaration,
				mesh.sharedVertexData->vertexBufferBinding,
				vertices,indices);
		}
		else
		{
			FillMeshData(smesh->indexData->indexBuffer,
				smesh->vertexData->vertexDeclaration,
				smesh->vertexData->vertexBufferBinding,
				vertices,indices);

		}
		if(indices.size()>0)
		{
			//start tootle work

			const float* pViewpoints = NULL;
			unsigned int nViewpoints = 0;

			// *****************************************************************
			//   Optimize the mesh
			// *****************************************************************

			unsigned int nTriangles = (unsigned int) indices.size() / 3;
			unsigned int nVertices = (unsigned int) vertices.size();
			const float* pVB = (float*) &vertices[0];
			unsigned int* pIB = (unsigned int*) &indices[0];
			unsigned int nStride = 3*sizeof(float);   

			TootleStats stats;
			TootleResult result;

			// initialize Tootle
			result = TootleInit();
			if( result != TOOTLE_OK )
			{
				DisplayTootleErrorMessage( result );
				return 1;
			}


			// measure input VCache efficiency
			result = TootleMeasureCacheEfficiency( pIB, nTriangles, settings.nCacheSize, &stats.fVCacheIn );
			if( result != TOOTLE_OK )
			{
				DisplayTootleErrorMessage( result );
				return 1;
			}

			// measure input overdraw.  Note that we assume counter-clockwise vertex winding. 
			result = TootleMeasureOverdraw( pVB, pIB, nVertices, nTriangles, nStride, pViewpoints, nViewpoints, settings.eWinding, 
				&stats.fOverdrawIn, &stats.fMaxOverdrawIn );
			if( result != TOOTLE_OK )
			{
				DisplayTootleErrorMessage( result );
				return 1;
			}

			// allocate an array to hold the cluster ID for each face
			std::vector<unsigned int> faceClusters;
			faceClusters.resize( nTriangles );

			// cluster the mesh, and sort faces by cluster
			result = TootleClusterMesh( pVB, pIB, nVertices, nTriangles, nStride, settings.nClustering, pIB, &faceClusters[0], NULL );
			if( result != TOOTLE_OK )
			{
				DisplayTootleErrorMessage( result );
				return 1;
			}

			stats.nClusters = 1 + (unsigned int)(faceClusters[ nTriangles - 1 ]);

			// perform vertex cache optimization on the clustered mesh
			result = TootleVCacheClusters( pIB, nTriangles, nVertices, settings.nCacheSize, &faceClusters[0], pIB, NULL );
			if( result != TOOTLE_OK )
			{
				DisplayTootleErrorMessage( result );
				return 1;
			}

			// optimize the draw order
			result = TootleOptimizeOverdraw( pVB, pIB, nVertices, nTriangles, nStride, pViewpoints, nViewpoints, settings.eWinding,
				&faceClusters[0], pIB, NULL );
			if( result != TOOTLE_OK )
			{
				DisplayTootleErrorMessage( result );
				return 1;
			}


			// measure output VCache efficiency
			result = TootleMeasureCacheEfficiency( pIB, nTriangles, settings.nCacheSize, &stats.fVCacheOut );
			if( result != TOOTLE_OK )
			{
				DisplayTootleErrorMessage( result );
				return 1;
			}

			// measure output overdraw
			result = TootleMeasureOverdraw( pVB, pIB, nVertices, nTriangles, nStride, pViewpoints, nViewpoints, settings.eWinding, 
				&stats.fOverdrawOut, &stats.fMaxOverdrawOut );
			if( result != TOOTLE_OK )
			{
				DisplayTootleErrorMessage( result );
				return 1;
			}

			// clean up tootle
			TootleCleanup();

			// print tootle statistics to stdout and stderr
			PrintStats(&stats ); 
			PrintStats(&stats ); 


			//copy the index buffer back to where it came from
			if (smesh->indexData->indexBuffer->getType() == HardwareIndexBuffer::IT_32BIT)
			{
				uint32	*pVIndices32 = NULL;    // the face indices buffer

				pVIndices32 = static_cast<uint32*>(
					smesh->indexData->indexBuffer->lock(HardwareBuffer::HBL_NORMAL));
				for(size_t i=0;i<smesh->indexData->indexBuffer->getNumIndexes();i++)
				{
					(*pVIndices32)=(uint32)indices[i];
					pVIndices32++;
				}

			}
			else
			{
				uint16	*pVIndices16 = NULL;    // the face indices buffer
				pVIndices16 = static_cast<uint16*>(
					smesh->indexData->indexBuffer->lock(HardwareBuffer::HBL_NORMAL));
				size_t nm_indices=smesh->indexData->indexBuffer->getNumIndexes();
				for(size_t i=0;i<nm_indices;i++)
				{
					(*pVIndices16)=(uint16)indices[i];
					pVIndices16++;
				}
			}
			smesh->indexData->indexBuffer->unlock();
			//clear the buffers for the next use
			vertices.clear();
			indices.clear();
		}
	}


	// save the modified Mesh file
	if(settings.mOutFile!=NULL)
	{
		mMeshSerializer.exportMesh(&mesh,settings.mOutFile);
		logMgr->getDefaultLog()->logMessage("Output Mesh Written Successfully.");
	}
	else
	{
		logMgr->getDefaultLog()->logMessage("No Output Parameter. No File Written");
	}


	delete meshMgr;
	delete skeletonSerializer;
	delete meshSerializer;
	delete skelMgr;
	delete matMgr;
	delete mth;
	delete rgm;
	delete logMgr;
	return 0;
}

